<?php

// This is a IA (chatgpt) powered conversational Bot ussing the Assistant API.
// You must supply your OpenAI API key and an assistant id for it to work. 
// You must create an assistant into OpenAI however you like. To get the 
// Assistant ID you can test your assistant in the Playground and you will 
// see the id (asst_xxxxxx) below the name.
//
// There are two special comands that the AI can use to handle conversations:
// [transfer] and [close]. You can add the following sentence to the assistnat
// system prompt to handle those cases:
//
// "If the user is angry, frustated or whats to talk to a real person, respond 
// saying that you will pass the conversation to a human and add the command 
// [transfer] in your reply, when finishing the conversation add the command
// [close]."
//
// There is also a GUARDRAILS_USER_PROMPT that you can use to append anything to 
// the user question/prompt. It can be used to restrict answers further.

$DEBUG        = 0;
$OPENAI_TOKEN = '';
$ASSISTANT_ID = '';

$GUARDRAILS_USER_PROMPT = '';

if($OPENAI_TOKEN=='' || $ASSISTANT_ID=='') {
    $response = "Bot is not yet configured, be sure to set up your OpenAI API key";
    send_response($response); // dies after sending response
}

$json = file_get_contents('php://input');
$data = json_decode($json,1);

log_debug("POST RECEIVED:\n".print_r($data,1));

$current_state = '';

// For assistants, the bot state is the openai thread
if($data['state']=='') {
    // No state, its a new conversation, create a thread
    $thread = openai_create_thread($OPENAI_TOKEN);
    $current_state = array("thread"=>$thread);
} else {
    $bot_state = json_decode($data['state'],1);
    $thread = $bot_state['thread'];
}

if($thread=='' || $thread === null) {
    $response = "Digital brain is not working correctly (A01) [transfer]";
    send_response($response); // dies after sending response
}

$user_prompt = $data['body'];
$user_prompt.=$GUARDRAILS_USER_PROMPT; // add guardrails to question

$msgid = openai_create_message($OPENAI_TOKEN,$thread,$user_prompt);

if($msgid=='' || $msgid === null) {
    $response = "Digital brain is not working correctly (A02) [transfer]";
    send_response($response); // dies after sending response
}

$runid = openai_create_run($OPENAI_TOKEN,$ASSISTANT_ID,$thread);
if($runid=='' || $runid === null) {
    $response = "Digital brain is not working correctly (A03) [transfer]";
    send_response($response); // dies after sending response
}

// Get run status up to 30 attempts/seconds
$ok=0;
for($a=1;$a<30;$a++) {
    sleep(1);
    $status = openai_retrieve_run($OPENAI_TOKEN,$thread,$runid);
    if($status=='completed') { $ok=1; break; }
}

if($ok==0) {
    $response = "Digital brain is not working correctly (A04) [transfer]";
    send_response($response); // dies after sending response
}

// Retrieve first one from list (desc order limit 2)
$messages = openai_list_messages($OPENAI_TOKEN,$thread);
$openai_assistant_reply =  $messages[0]['content'][0]['text']['value'];

$command = extract_command_from_reply($openai_assistant_reply);
if($command=='close' || $command=='transfer') {
    openai_delete_thread($OPEANI_TOKEN,$thread);
}

$final_msg = remove_annotations($openai_assistant_reply);
send_response($final_msg);
// END OF PROGRAM

function openai_create_thread($openai_token) {
    $url = 'https://api.openai.com/v1/threads';
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v1");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_POST, 1);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create thread response: ".print_r($response,1));
    return $response['id'];
}

function openai_delete_thread($openai_token,$threadid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid;
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v1");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "DELETE");
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("delete trhead response: ".print_r($response,1));
    return $response['deleted'];
}

function openai_create_message($openai_token,$threadid,$content) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/messages";
    $data = array(
       "role"=> "user",
       "content" => $content
    );
    $data_string = json_encode($data);
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v1");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create message url: ".$url);
    log_debug("create message payload: ". $data_String);
    log_debug("create message response: ".print_r($response,1));
    return $response['id'];
}

function openai_list_messages($openai_token,$threadid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/messages?order=desc&limit=2";
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v1");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create message response: ".print_r($response,1));
    return $response['data'];
}

function openai_create_run($openai_token,$assistantid,$threadid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/runs";
    $data = array(
       "assistant_id"=> $assistantid
    );
    $data_string = json_encode($data);
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v1");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_POST, 1);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create run response: ".print_r($response,1));
    return $response['id'];
}

function openai_retrieve_run($openai_token,$threadid,$runid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/runs/".$runid;
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v1");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("retrieve run response: ".print_r($response,1));
    return $response['status'];
}

function extract_command_from_reply($text) {
    $pattern = '/\[(.*?)\]/';
    preg_match($pattern, $text, $matches);
    $extractedString = (isset($matches[1])) ? $matches[1] : '';
    return $extractedString;
}

function remove_annotations($text) {
    $pattern = '/【(.*?)】/';
    $replacedString = preg_replace($pattern, '', $text);
    return $replacedString;
}

function send_response($response) {
    global $current_state;
    $return   = array("body"=>$response,"attach"=>'');
    if($current_state!='') {
        $return['state']=$current_state;
    }
    log_debug("REPLY: ".print_r($return,1));
    echo json_encode($return);
    die();
}

function log_debug($text) {
    global $DEBUG;
    $logfile = basename(__FILE__,'.php');
    if($DEBUG==1) {
        file_put_contents("/tmp/${logfile}.log",$text."\n",FILE_APPEND);
    }
}
